/*
  averages.hpp

  Copyright (C) 2015 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _AVERAGES_HPP_
#define _AVERAGES_HPP_
#include <numeric>
#include <algorithm>
#include <cmath>

#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
#define INTERNAL_VALUE_TYPE_OF_ITERATOR long double
#else
#define INTERNAL_VALUE_TYPE_OF_ITERATOR typename std::iterator_traits<InputIterator>::value_type
#endif

template <typename InputIterator>
typename std::iterator_traits<InputIterator>::value_type arithmetic_average(InputIterator first, InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
#if 0
  return std::accumulate(first, last, value_type(0))/std::distance(first, last);
#elif 0
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last)
    v += *it++/std::distance(first, last);
  return v;
#else
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last)
    v += (*it - v)/std::distance(first, ++it);
  return v;
#endif
}

template <typename InputIterator>
typename std::iterator_traits<InputIterator>::value_type geometric_average(InputIterator first, InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
#if 0
  return pow(std::accumulate(first, last, value_type(1), std::multiplies<value_type>()), value_type(1)/std::distance(first, last));
#elif 0
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last)
    v += log(*it++)/std::distance(first, last);
  return exp(v);
#else
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last)
    v += (log(*it) - v)/std::distance(first, ++it);
  return exp(v);
#endif
}

template <typename InputIterator>
typename std::iterator_traits<InputIterator>::value_type harmonic_average(InputIterator first, InputIterator last)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last)
    v += value_type(1)/(*it++);
  return std::distance(first, last)/v;
}

template <typename InputIterator>
typename std::iterator_traits<InputIterator>::value_type generalized_average(InputIterator first, InputIterator last, typename std::iterator_traits<InputIterator>::value_type power = 2)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
#if 0
  while (it != last)
    v += pow(*it++, power);
  return pow(v/std::distance(first, last), value_type(1)/power);
#elif 0
  while (it != last)
    v += pow(*it++, power)/std::distance(first, last);
  return pow(v, value_type(1)/power);
#else
  while (it != last)
    v += (pow(*it, power) - v)/std::distance(first, ++it);
  return pow(v, value_type(1)/power);
#endif
}

template <typename InputIterator, typename UnaryOperation, typename InverseUnaryOperation>
typename std::iterator_traits<InputIterator>::value_type generalized_average(InputIterator first, InputIterator last, UnaryOperation unary_operation, InverseUnaryOperation inverse_unary_operation)
{
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
#if 0
  while (it != last)
    v += unary_operation(*it++);
  return inverse_unary_operation(v/std::distance(first, last));
#elif 0
  while (it != last)
    v += unary_operation(*it++)/std::distance(first, last);
  return inverse_unary_operation(v);
#else
  while (it != last)
    v += (unary_operation(*it) - v)/std::distance(first, ++it);
  return inverse_unary_operation(v);
#endif
}

template <typename InputIterator>
typename std::iterator_traits<InputIterator>::value_type exponential_smoothed_average(InputIterator first, InputIterator last, typename std::iterator_traits<InputIterator>::value_type smoothing_factor = .1)
{
  typedef typename std::iterator_traits<InputIterator>::value_type value_type;
  InputIterator it(first);
  INTERNAL_VALUE_TYPE_OF_ITERATOR v(0);
  while (it != last)
    v = v*(value_type(1) - smoothing_factor) + *it++*smoothing_factor;
  return v;
}

#endif
