/*
  histogram.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _HISTOGRAM_HPP_
#define _HISTOGRAM_HPP_

#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
#define INTERNAL_VALUE_TYPE long double
#else
#define INTERNAL_VALUE_TYPE T
#endif

template <typename InputIterator>
size_t histogram(std::vector<size_t> &heights,
		 const InputIterator first, const InputIterator last,
		 const typename std::iterator_traits<InputIterator>::value_type lb,
		 const typename std::iterator_traits<InputIterator>::value_type ub)
{
  size_t rv(0);
  const typename std::iterator_traits<InputIterator>::value_type sz(ub - lb);
  for (InputIterator it(first); it!=last; ++it) {
    if (*it < lb) continue;
    const typename std::iterator_traits<InputIterator>::value_type v(*it - lb);
    const size_t id(v/sz*heights.size());
    if (!(id < heights.size())) continue;
    ++heights[id];
    ++rv;
  }
  return rv;
}
template <typename T>
inline
T histogram_lower_bound(const size_t i, const size_t n_bins, const T lb, const T ub, const bool option_lt_ub = false)
{
  if (option_lt_ub)
    return lb + (ub - lb)/n_bins*i;
  else
    return T(n_bins - i)/n_bins*lb + T(i)/n_bins*ub;
}
template <typename T>
inline
T histogram_upper_bound(const size_t i, const size_t n_bins, const T lb, const T ub, const bool option_lt_ub = false)
{
  return histogram_lower_bound(i+1, n_bins, lb, ub, option_lt_ub);
}
template <typename T>
inline
std::pair<T, T> histogram_range(const size_t i, const size_t n_bins, const T lb, const T ub, const bool option_lt_ub = false)
{
  return std::make_pair(histogram_lower_bound(i, n_bins, lb, ub, option_lt_ub),
			histogram_upper_bound(i, n_bins, lb, ub, option_lt_ub));
}
template <typename T>
T histogram_mean(const std::vector<size_t> &heights, const T lb, const T ub)
{
  INTERNAL_VALUE_TYPE w(0), v(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!(*it > 0)) continue;
    const std::pair<T, T> range(histogram_range(std::distance(heights.begin(), it), heights.size(), lb, ub));
    const INTERNAL_VALUE_TYPE p((range.first + range.second)/2);
    w += *it;
    v += (p - v)*((INTERNAL_VALUE_TYPE)*it/w);
  }
  return v;
}
template <typename T>
T histogram_variance(const std::vector<size_t> &heights, const T lb, const T ub)
{
  INTERNAL_VALUE_TYPE mu(histogram_mean(heights, lb, ub)), w(0), v(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!(*it > 0)) continue;
    const std::pair<T, T> range(histogram_range(std::distance(heights.begin(), it), heights.size(), lb, ub));
    const INTERNAL_VALUE_TYPE p((range.first + range.second)/2), delta(p - mu);
    w += *it;
    v += (delta*delta - v)*((INTERNAL_VALUE_TYPE)*it/w);
  }
  return v;
}
template <typename T>
T histogram_standard_deviation(const std::vector<size_t> &heights, const T lb, const T ub)
{
  return sqrt(histogram_variance(heights, lb, ub));
}

template <typename InputIterator>
size_t histogram_logarithm(std::vector<size_t> &heights,
			   const InputIterator first, const InputIterator last,
			   const typename std::iterator_traits<InputIterator>::value_type lb,
			   const typename std::iterator_traits<InputIterator>::value_type ub)
{
  size_t rv(0);
  const typename std::iterator_traits<InputIterator>::value_type sz(log(ub) - log(lb));
  for (InputIterator it(first); it!=last; ++it) {
    if (*it < lb) continue;
    const typename std::iterator_traits<InputIterator>::value_type v(log(*it) - log(lb));
    const size_t id(v/sz*heights.size());
    if (!(id < heights.size())) continue;
    ++heights[id];
    ++rv;
  }
  return rv;
}
template <typename T>
inline
T histogram_logarithm_lower_bound(const size_t i, const size_t n_bins, const T lb, const T ub, const bool option_lt_ub = false)
{
#if 0
  if (option_lt_ub)
    return lb*exp((log(ub) - log(lb))*i/n_bins);
  else
    return exp((T)(n_bins - i)/n_bins*log(lb) + (T)i/n_bins*log(ub));
#else
  if (option_lt_ub)
    return lb*expl((logl((INTERNAL_VALUE_TYPE)ub) - logl((INTERNAL_VALUE_TYPE)lb))*i/n_bins);
  else
    return expl((INTERNAL_VALUE_TYPE)(n_bins - i)/n_bins*logl(lb) + (INTERNAL_VALUE_TYPE)i/n_bins*logl(ub));
#endif
}
template <typename T>
inline
T histogram_logarithm_upper_bound(const size_t i, const size_t n_bins, const T lb, const T ub, const bool option_lt_ub = false)
{
  return histogram_logarithm_lower_bound(i+1, n_bins, lb, ub, option_lt_ub);
}
template <typename T>
inline
std::pair<T, T> histogram_logarithm_range(const size_t i, const size_t n_bins, const T lb, const T ub, const bool option_lt_ub = false)
{
  return std::make_pair(histogram_logarithm_lower_bound(i, n_bins, lb, ub, option_lt_ub),
			histogram_logarithm_upper_bound(i, n_bins, lb, ub, option_lt_ub));
}
template <typename T>
T histogram_logarithm_mean(const std::vector<size_t> &heights, const T lb, const T ub)
{
  INTERNAL_VALUE_TYPE w(0), v(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!(*it > 0)) continue;
    const std::pair<T, T> range(histogram_logarithm_range(std::distance(heights.begin(), it), heights.size(), lb, ub));
#if 0
    const INTERNAL_VALUE_TYPE p((range.first + range.second)/2);
#else
    const INTERNAL_VALUE_TYPE p(exp((log(range.first) + log(range.second))/2));
#endif
    w += *it;
    v += (p - v)*((INTERNAL_VALUE_TYPE)*it/w);
  }
  return v;
}
template <typename T>
T histogram_logarithm_variance(const std::vector<size_t> &heights, const T lb, const T ub)
{
  INTERNAL_VALUE_TYPE mu(histogram_logarithm_mean(heights, lb, ub)), w(0), v(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!(*it > 0)) continue;
    const std::pair<T, T> range(histogram_logarithm_range(std::distance(heights.begin(), it), heights.size(), lb, ub));
#if 0
    const INTERNAL_VALUE_TYPE p((range.first + range.second)/2), delta(p - mu);
#else
    const INTERNAL_VALUE_TYPE p(exp((log(range.first) + log(range.second))/2)), delta(p - mu);
#endif
    w += *it;
    v += (delta*delta - v)*((INTERNAL_VALUE_TYPE)*it/w);
  }
  return v;
}
template <typename T>
T histogram_logarithm_standard_deviation(const std::vector<size_t> &heights, const T lb, const T ub)
{
  return sqrt(histogram_logarithm_variance(heights, lb, ub));
}

#endif
