/*
  statistics_online.hpp

  Copyright (C) 2015 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _STATISTICS_ONLINE_HPP_
#define _STATISTICS_ONLINE_HPP_
#include "averages_online.hpp"
#include "statistics.hpp"

template <typename T>
class statistics_online : public arithmetic_average_online<T> {
public:
  statistics_online() : arithmetic_average_online<T>(), harmonic_mean(0), positive_minimum_value(0), negative_maximum_value(0), sample_variance(0), logarithmic_mean(0), root_mean(0), Mu(0), Mu2(0), Mu3(0), Mu4(0) {}
  void add(const T v)
  {
    minimum_value = arithmetic_average_online<T>::get_count() ? std::min(T(minimum_value), v) : v;
    maximum_value = arithmetic_average_online<T>::get_count() ? std::max(T(maximum_value), v) : v;
    positive_minimum_value = (0 < v && 0 < positive_minimum_value) ? std::min(T(positive_minimum_value), v) : (0 < v) ? v : positive_minimum_value;
    negative_maximum_value = (v < 0 && negative_maximum_value < 0) ? std::max(T(negative_maximum_value), v) : (v < 0) ? v : negative_maximum_value;
    typename arithmetic_average_online<T>::value_type omu(arithmetic_average_online<T>::get_value());
    unsigned long long previous_count(arithmetic_average_online<T>::get_count());
    typename arithmetic_average_online<T>::value_type mu(arithmetic_average_online<T>::average(v));
    unsigned long long count(arithmetic_average_online<T>::get_count());
    logarithmic_mean += (log(v) - logarithmic_mean)/count;
    geometric_mean = exp(logarithmic_mean);
    harmonic_mean = previous_count ? count/(previous_count/harmonic_mean + typename arithmetic_average_online<T>::value_type(1)/v) : v;
    root_mean += (v*v - root_mean)/count;
    root_mean_square = sqrt(root_mean);
    sample_variance += ((v - omu)*(v - mu) - sample_variance)/count;
    typename generic_average_online<T>::value_type
      pn = previous_count, n = count,
      delta(v - Mu),
      delta_par_n(delta/n),
      square_delta_par_n(delta_par_n*delta_par_n),
      delta_times_delta_par_n_times_pn(delta*delta_par_n*pn);
    Mu4 += delta_times_delta_par_n_times_pn*square_delta_par_n*(n*n - 3*n + 3) + 6*square_delta_par_n*Mu2 - 4*delta_par_n*Mu3;
    Mu3 += delta_par_n*(delta_times_delta_par_n_times_pn*(pn - 1) - 3*Mu2);
    Mu2 += delta_times_delta_par_n_times_pn;
    Mu += delta_par_n;
    unbiased_variance = Mu2/pn;
    skewness = (pow(pn, 1.5)/n) * Mu3/pow(Mu2, 1.5);
    kurtosis = (pn/n*pn)*Mu4 / (Mu2*Mu2) - 3;
  }
  // restarts when 'count' will become zero again.
  T get_mean() const { return arithmetic_average_online<T>::get_value(); }
  T get_geometric_mean() const { return geometric_mean; }
  T get_harmonic_mean() const { return harmonic_mean; }
  T get_root_mean_square() const { return root_mean_square; }
  T get_sample_variance() const { return sample_variance; }
  T get_sample_standard_deviation() const { return sqrt(sample_variance); }
  T get_unbiased_variance() const { return unbiased_variance; }
  T get_unbiased_standard_deviation() const { return sqrt(unbiased_variance); }
  T get_third_moment() const { return Mu3/arithmetic_average_online<T>::get_count(); };
  T get_skewness() const { return skewness; };
  T get_fourth_moment() const { return Mu4/arithmetic_average_online<T>::get_count(); };
  T get_kurtosis() const { return kurtosis; };
  T get_minimum_value() const { return minimum_value; }
  T get_maximum_value() const { return maximum_value; }
  T get_positive_minimum_value() const { return positive_minimum_value; }
  T get_negative_maximum_value() const { return negative_maximum_value; }
  void reset() { arithmetic_average_online<T>::reset(); positive_minimum_value = 0; negative_maximum_value = 0; sample_variance = 0; logarithmic_mean = 0; root_mean = 0; Mu = Mu2 = Mu3 = Mu4 = 0; }
protected:
  typename arithmetic_average_online<T>::value_type geometric_mean, harmonic_mean, root_mean_square;
  typename arithmetic_average_online<T>::value_type unbiased_variance, skewness, kurtosis;
  typename arithmetic_average_online<T>::value_type positive_minimum_value, negative_maximum_value, minimum_value, maximum_value, sample_variance;
  typename arithmetic_average_online<T>::value_type logarithmic_mean, root_mean;
  typename arithmetic_average_online<T>::value_type Mu, Mu2, Mu3, Mu4;
};

template <typename T>
class moving_statistics_online : public moving_arithmetic_average_online<T>, public statistics_online<T> {
public:
  moving_statistics_online(const size_t size = 100) : moving_arithmetic_average_online<T>(size), statistics_online<T>() {}
  void add(const typename generic_average_online<T>::value_type v)
  {
    const size_t size(moving_arithmetic_average_online<T>::size);
    const size_t index(moving_arithmetic_average_online<T>::get_index());
    const T *storage(moving_arithmetic_average_online<T>::get_storage());
    const typename generic_average_online<T>::value_type ov(storage[index]);
    typename generic_average_online<T>::value_type omu(get_mean());
    bool initial(moving_arithmetic_average_online<T>::is_initial());
    typename generic_average_online<T>::value_type mu(moving_arithmetic_average_online<T>::average(v));
    if (initial)
      statistics_online<T>::add(v);
    else {
      statistics_online<T>::logarithmic_mean += (log(v) - log(ov))/size;
      statistics_online<T>::geometric_mean = exp(statistics_online<T>::logarithmic_mean);

      statistics_online<T>::harmonic_mean = typename statistics_online<T>::value_type(1)/statistics_online<T>::harmonic_mean;
      statistics_online<T>::harmonic_mean += (typename statistics_online<T>::value_type(1)/v - typename statistics_online<T>::value_type(1)/ov)/size;
      statistics_online<T>::harmonic_mean = typename statistics_online<T>::value_type(1)/statistics_online<T>::harmonic_mean;

      statistics_online<T>::root_mean += (v + ov)*(v - ov)/size;
      statistics_online<T>::root_mean_square = sqrt(statistics_online<T>::root_mean);

      omu = statistics_online<T>::Mu;
      statistics_online<T>::Mu += (v - ov)/size;
      mu = statistics_online<T>::Mu;

      statistics_online<T>::sample_variance += (v + ov)*(v - ov)/size + (omu + mu)*(omu - mu);

      const typename statistics_online<T>::value_type oMu2(statistics_online<T>::Mu2);
      statistics_online<T>::Mu2 += (v + ov)*(v - ov) + (omu + mu)*(omu - mu)*size;
      statistics_online<T>::unbiased_variance = statistics_online<T>::Mu2/(size - 1);

      const typename statistics_online<T>::value_type oMu3(statistics_online<T>::Mu3);
#undef DO_NOT_USE_MOVING_SKEWNESS_ONLINE
#if !defined(DO_NOT_USE_MOVING_SKEWNESS_ONLINE)
      statistics_online<T>::Mu3 += (v*v + v*ov + ov*ov)*(v - ov) - 3*(mu*statistics_online<T>::Mu2 - omu*oMu2) - (mu*mu + mu*omu + omu*omu)*(mu - omu)*size;
      statistics_online<T>::skewness = (pow((size - 1), 1.5)/size) * statistics_online<T>::Mu3/pow(statistics_online<T>::Mu2, 1.5);
#else
      statistics_online<T>::Mu3 = third_moment(&storage[0], &storage[size])*size;
#endif
#undef DO_NOT_USE_MOVING_KURTOSIS_ONLINE
#if !defined(DO_NOT_USE_MOVING_KURTOSIS_ONLINE)
      statistics_online<T>::Mu4 += (v*v + ov*ov)*(v + ov)*(v - ov) - 4*(mu*statistics_online<T>::Mu3 - omu*oMu3) - 6*(mu*mu*statistics_online<T>::Mu2 - omu*omu*oMu2) - (mu*mu + omu*omu)*(mu + omu)*(mu - omu)*size;
      statistics_online<T>::kurtosis = (typename statistics_online<T>::value_type(size - 1)/size*(size - 1))*statistics_online<T>::Mu4 / (statistics_online<T>::Mu2*statistics_online<T>::Mu2) - 3;
#else
      statistics_online<T>::Mu4 = fourth_moment(&storage[0], &storage[size])*size;
#endif
      statistics_online<T>::minimum_value =
	(ov == statistics_online<T>::minimum_value) ?
	((v < ov) ? v : *std::min_element(&storage[0], &storage[size])) :
	*std::min_element(&storage[0], &storage[size]);

      statistics_online<T>::maximum_value =
	(ov == statistics_online<T>::maximum_value) ?
	((ov < v) ? v : *std::max_element(&storage[0], &storage[size])) :
	*std::max_element(&storage[0], &storage[size]);

      statistics_online<T>::positive_minimum_value =
	(ov == statistics_online<T>::positive_minimum_value) ?
	((0 < v && (ov == 0 || v < ov)) ? v : positive_min_value(&storage[0], &storage[size])) :
	positive_min_value(&storage[0], &storage[size]);

      statistics_online<T>::negative_maximum_value =
	(ov == statistics_online<T>::negative_maximum_value) ?
	((v < 0 && (ov == 0 || ov < v)) ? v : negative_max_value(&storage[0], &storage[size])) :
	negative_max_value(&storage[0], &storage[size]);
    }
  }
  T get_mean() const { return moving_arithmetic_average_online<T>::get_value(); }
#if !defined(DO_NOT_USE_MOVING_SKEWNESS_ONLINE)
#else
  T get_skewness()
  {
    const size_t size(moving_arithmetic_average_online<T>::size);
    const T *storage(moving_arithmetic_average_online<T>::get_storage());
    if (!moving_arithmetic_average_online<T>::is_initial())
      statistics_online<T>::skewness = skewness(&storage[0], &storage[size]);
    return statistics_online<T>::get_skewness();
  }
#endif
#if !defined(DO_NOT_USE_MOVING_KURTOSIS_ONLINE)
#else
  T get_kurtosis()
  {
    const size_t size(moving_arithmetic_average_online<T>::size);
    const T *storage(moving_arithmetic_average_online<T>::get_storage());
    if (!moving_arithmetic_average_online<T>::is_initial())
      statistics_online<T>::kurtosis = kurtosis(&storage[0], &storage[size]);
    return statistics_online<T>::get_kurtosis();
  }
#endif
  void reset() { statistics_online<T>::reset(); moving_arithmetic_average_online<T>::reset(); }
};

#endif
